// BSP functions    by Alan Baylis 2001

#include "shared.h"
#include "listnode.h"
#include "bsp.h"
#include "locmath.h"
#include "polygon.h"
#include "camera.h"
#include "collision.h"
#include "lightmap.h"
#include "tll.h"
#include "bullet.h"
#include "particle.h"


extern CAMERA* camera;
extern int currentCamera;
extern LinkedList<BSP_node> LeafList;
extern LinkedList<BSP_node> PartitionList;
extern BSP_node* listnode;
extern int numcurrentportals;
extern int numleaves;
extern int numpartitions;
extern int showportals;
extern int currentleaf;
extern int SphereSector;
extern VECTOR SpherePosition;
extern int numBullets;
extern BULLET* bullet;

int BSP_node::Compare(const BSP_node& Node)
{
  if (linkPosition < Node.linkPosition)
    return smaller;
  if (linkPosition > Node.linkPosition)
    return bigger;
  else
    return same;
}

void MakeNodeLists(BSP_node* node)
{
    if (node->leaf == true)
    {
        node->linkPosition = ++LeafList.numObjects;;
        LeafList.Insert(node);
        return;
    }
    else
    {
        if (node->nodeid != 0)
        {
            node->linkPosition = ++PartitionList.numObjects;
            PartitionList.Insert(node);
        }
    }
    MakeNodeLists(node->frontnode);
    MakeNodeLists(node->backnode);
}

int FindCurrentLeaf(VECTOR Position, BSP_node* node)
{
    if (node->leaf == true)
    {
        numcurrentportals = node->numportals;
        return node->nodeid;
    }

    VECTOR edge1, edge2, planeNormal, temp;
    // get the partitioning planes normal
    edge1 = node->partition.Vertex[1].coords - node->partition.Vertex[0].coords;
    edge2 = node->partition.Vertex[2].coords - node->partition.Vertex[0].coords;
    planeNormal = CrossVector(edge1, edge2);
    temp = node->partition.Vertex[0].coords;

    int side = ClassifyPoint(Position, temp, planeNormal);

    if (side == 1 || side == 0)
    {
        return FindCurrentLeaf(Position, node->frontnode);
    }
    else
    {
        return FindCurrentLeaf(Position, node->backnode);
    }
}

int SelectPartitionfromList(POLYGON* nodepolylist, int numpolys, int* bestfront, int* bestback)
{
    int count = 0, result, absdifference = 1000000000, bestplane = 0, front, back, potentialplane, polytoclassify;
    VECTOR temp;

    // Loop through all the polygons and find the best splitting plane
    for(potentialplane = 0; potentialplane < numpolys; potentialplane++)
    {
        front = back = 0;
        for (polytoclassify = 0; polytoclassify < numpolys; polytoclassify++)
        {
            result = SplitTriangle(nodepolylist[polytoclassify], nodepolylist[potentialplane], NULL);
            switch (result)
            {
                case Front:
                    front++;
                break;

                case Back:
                    back++;
                break;

                case TwoFrontOneBack:
                    front += 2;
                    back++;
                break;

                case OneFrontTwoBack:
                    front++;
                    back += 2;
                break;

                case OneFrontOneBack:
                    front++;
                    back++;
                break;
            }
        }
        if (abs(front - back) < absdifference)
        {
            absdifference = abs(front - back);
            bestplane = potentialplane;
            *bestfront = front;
            *bestback = back;
        }
        if (front == 0 || back == 0)
            count++;
    }
    if (count == numpolys)
        return -1;
    else
        return bestplane;
}

void BuildBSP(BSP_node* node)
{
    int result, front, back, polytoclassify, partplane;
    POLYGON output[3];

    partplane = SelectPartitionfromList(node->nodepolylist, node->numpolys, &front, &back);

    if (partplane == -1)
    {
        node->nodeid = ++numleaves;
        node->leaf = true;
        return;
    }

    node->nodeid = ++numpartitions;
    node->partition = node->nodepolylist[partplane];

    //Allocate memory for a front and back node
    node->frontnode = new BSP_node;
    node->frontnode->visible = 0;
    node->frontnode->leaf = 0;
    node->frontnode->numpolys = front;
    node->frontnode->nodepolylist = new POLYGON[front];
    node->frontnode->numportals = 0;
    node->frontnode->numdecals = 0;

    node->backnode = new BSP_node;
    node->backnode->visible = 0;
    node->backnode->leaf = 0;
    node->backnode->numpolys = back;
    node->backnode->nodepolylist = new POLYGON[back];
    node->backnode->numportals = 0;
    node->backnode->numdecals = 0;

    //Classify each polygon in the current node with respect to the partitioning plane.
    front = back = 0;
    for (polytoclassify = 0; polytoclassify < node->numpolys; polytoclassify++)
    {
        output[0] = node->nodepolylist[polytoclassify];
        output[1] = node->nodepolylist[polytoclassify];
        output[2] = node->nodepolylist[polytoclassify];

        result = SplitTriangle(node->nodepolylist[polytoclassify], node->partition, output);
        switch (result)
        {
            case Front:
                node->frontnode->nodepolylist[front] = node->nodepolylist[polytoclassify];
                front++;
            break;

            case Back:
                node->backnode->nodepolylist[back] = node->nodepolylist[polytoclassify];
                back++;
            break;

            case TwoFrontOneBack:
                node->frontnode->nodepolylist[front] = output[0];
                node->frontnode->nodepolylist[front + 1] = output[1];
                front += 2;
                node->backnode->nodepolylist[back] = output[2];
                back++;
            break;

            case OneFrontTwoBack:
                node->frontnode->nodepolylist[front] = output[0];
                front++;
                node->backnode->nodepolylist[back] = output[1];
                node->backnode->nodepolylist[back + 1] = output[2];
                back += 2;
            break;

            case OneFrontOneBack:
                node->frontnode->nodepolylist[front] = output[0];
                front++;
                node->backnode->nodepolylist[back] = output[1];
                back++;
            break;
        }
    }

    node->numpolys = 0;
    delete[] node->nodepolylist;

    BuildBSP(node->frontnode);
    BuildBSP(node->backnode);
}

int RenderBSP(BSP_node* node)
{
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

    int Side;
    VECTOR Position, edge1, edge2, planeNormal, temp;

    //The current position of the player/viewpoint
    Position = camera[currentCamera].Position;

    if (!node->leaf)
    {
        // get the partitioning planes normal
        edge1 = node->partition.Vertex[1].coords - node->partition.Vertex[0].coords;
        edge2 = node->partition.Vertex[2].coords - node->partition.Vertex[0].coords;
        planeNormal = CrossVector(edge1, edge2);
        temp = node->partition.Vertex[0].coords;
        Side = ClassifyPoint(Position, temp, planeNormal);

        if (Side == -1)
        {
            RenderBSP(node->frontnode);
            RenderBSP(node->backnode);
        }
        else
        {
            RenderBSP(node->backnode);
            RenderBSP(node->frontnode);
        }
    }

    if (node->leaf && node->visible)
    {
        node->visible = false;

        //Draw polygons that are in the leaf
        for (int loop = 0; loop < node->numpolys; loop++)
        {
            glMatrixMode(GL_TEXTURE);
            glPushMatrix();

            glScalef(node->nodepolylist[loop].Scale[0], node->nodepolylist[loop].Scale[1], 1.0f);
            glTranslatef(node->nodepolylist[loop].Shift[0], node->nodepolylist[loop].Shift[1], 0.0f);
            glRotatef(node->nodepolylist[loop].Rotate, 0.0f, 0.0f, 1.0f);
            glBindTexture(GL_TEXTURE_2D, node->nodepolylist[loop].Texture);
            glBegin(GL_TRIANGLES);
                glNormal3fv(&node->nodepolylist[loop].Vertex[0].normal.x);
                glTexCoord2f(node->nodepolylist[loop].Vertex[0].u, node->nodepolylist[loop].Vertex[0].v);
                glVertex3fv(&node->nodepolylist[loop].Vertex[0].coords.x);
                glTexCoord2f(node->nodepolylist[loop].Vertex[1].u, node->nodepolylist[loop].Vertex[1].v);
                glVertex3fv(&node->nodepolylist[loop].Vertex[1].coords.x);
                glTexCoord2f(node->nodepolylist[loop].Vertex[2].u, node->nodepolylist[loop].Vertex[2].v);
                glVertex3fv(&node->nodepolylist[loop].Vertex[2].coords.x);
            glEnd();
            glPopMatrix();
            glMatrixMode(GL_MODELVIEW);

            glEnable(GL_BLEND);
            glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR);

            glBindTexture(GL_TEXTURE_2D, node->nodelightmaplist[loop].Texture.TexID);
            glBegin(GL_TRIANGLES);
                glNormal3fv(&node->nodepolylist[loop].Vertex[0].normal.x);
                glTexCoord2f(node->nodelightmaplist[loop].vertex_u[0], node->nodelightmaplist[loop].vertex_v[0]);
                glVertex3fv(&node->nodepolylist[loop].Vertex[0].coords.x);
                glTexCoord2f(node->nodelightmaplist[loop].vertex_u[1], node->nodelightmaplist[loop].vertex_v[1]);
                glVertex3fv(&node->nodepolylist[loop].Vertex[1].coords.x);
                glTexCoord2f(node->nodelightmaplist[loop].vertex_u[2], node->nodelightmaplist[loop].vertex_v[2]);
                glVertex3fv(&node->nodepolylist[loop].Vertex[2].coords.x);
            glEnd();
            glDisable(GL_BLEND);
        }
/*
        if (showportals)
        {
            // Draw the leaf portals
            glDisable(GL_TEXTURE_2D);
            glDisable(GL_LIGHTING);
            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            glEnable(GL_ALPHA_TEST);
            glAlphaFunc(GL_GREATER, 0);
            glColor4f(1.0, 1.0, 0.0, 0.2);

            if (currentleaf == node->nodeid)
            {
                for (int loop = 1; loop <= node->numportals; loop++)
                {
                    PORTAL* tempportal = node->portallist.Get(loop);
                    glBegin(GL_POLYGON);
                        glNormal3fv(&tempportal->Vertex[0].normal.x);
                        for (int innerloop = 0; innerloop < tempportal->numVertices; innerloop++)
                            glVertex3fv(&tempportal->Vertex[innerloop].coords.x);
                    glEnd();
                }
            }
            glDisable(GL_BLEND);
            glDisable(GL_ALPHA_TEST);
            glEnable(GL_TEXTURE_2D);
            glEnable(GL_LIGHTING);
        }
*/
        // Draw decals
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        DECAL* tempDecal;
        for (int loop = 1; loop <= node->numdecals; loop++)
        {
            tempDecal = node->decallist.Get(loop);
            if (tempDecal->type == 0)
                RenderBulletDecal(*tempDecal);
            else
                RenderBurnDecal(*tempDecal);
            --tempDecal->counter;
            if (tempDecal->counter == 0)
            {
                node->decallist.Delete(loop);
                --node->numdecals;
            }
        }

        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

        // Draw bullets
        for (int loop = 0; loop < numBullets; loop++)
        {
            if (bullet[loop].leaf == node->nodeid && bullet[loop].active)
                bullet[loop].Draw();
        }

        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        // Draw particles
        RenderParticles(node->nodeid);
    }
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    return 1;
}

void DeleteBSP(BSP_node *node)
{
    if (node->leaf == true)
    {
        delete[] node->nodepolylist;
        delete[] node->nodelightmaplist;
        for (int i = node->numportals; i > 0 ; i--)
        {
            PORTAL* temp = node->portallist.Get(i);
            delete[] temp->Vertex;
            node->portallist.Delete(i);
            delete temp;
        }
        node->numportals = 0;
        return;
    }

    DeleteBSP(node->frontnode);
    delete node->frontnode;
    DeleteBSP(node->backnode);
    delete node->backnode;
}

void DrawIntersectionSphere(VECTOR coordinates)
{
    float mat_ambient[] = { 0.2, 1.0, 0.1, 1.0 };
    float mat_diffuse[] = { 0.2, 1.0, 0.1, 1.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glDisable(GL_TEXTURE_2D);
    glPushMatrix();
    glTranslatef(coordinates.x, coordinates.y, coordinates.z);
    GLUquadricObj * sphere = gluNewQuadric();
    gluQuadricOrientation(sphere, GLU_OUTSIDE);
    gluSphere(sphere,0.3,20,20);
    glPopMatrix();
    glEnable(GL_TEXTURE_2D);
}
