// ~~~ OpenGL & Win32 ~~~
// Tutorial No.20 - Constructive Solid Geometry
// Alan Baylis 21/03/2003

#include <windows.h>
#include "shared.h"
#include "general.h"
#include "listnode.h"
#include "locmath.h"
#include "vector.h"
#include "vertex.h"
#include "quat.h"
#include "matrix.h"
#include "texture.h"
#include "polygon.h"
#include "tll.h"
#include "plane.h"
#include "object.h"
#include "camera.h"
#include "light.h"
#include "collision.h"
#include "glfont.h"
#include "bspline.h"
#include "bsp.h"
#include "portal.h"
#include "pvs.h"
#include "tga.h"
#include "lightmap.h"
#include "winfuncs.h"
#include "console.h"
#include "log.h"
#include "bass.h"
#include "sound.h"
#include "bullet.h"
#include "decal.h"
#include "particle.h"
#include "brush.h"
#include "polygon.h"
#include "csg.h"
#include "resource.h"

// Console
ConsoleWindow Console;

// Windows
HGLRC hRC;
HDC hDC;
HWND hWnd;
HINSTANCE g_hInst;
RECT screen;
PAINTSTRUCT ps;
char AppDirectory[MAX_PATH];
float ApplicationStartTime;

// Math
float pi = 3.141592;
float radian = pi / 180;
float epsilon = 0.05;

// Input
bool key[256];
bool released_key[256];

// Cameras
int currentCamera = 0;
int numCameras = 4;
float step = 30.0;

// Timing
float lasttime;
float multiplier;
const int average = 10;
float lastmultiplier[average];
int fps;

// Lights
int GL_LIGHT[8] = {GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3, GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7};
int currentLight = 0;
int numLights = 0;

// View Frustum Planes
PLANE frustum[6];

// World vertices
int numVertices = 37;
VERTEX* vertex = new VERTEX[numVertices];

// World polygons (these are in fact triangles)
int numWorldPolygons = 58;
POLYGON* worldPolygons = new POLYGON[numWorldPolygons];
POLYGON* bspPolygons = new POLYGON[numWorldPolygons];

// Texture
TEXTURE* texture = new TEXTURE[11];

// Camera
CAMERA* camera = new CAMERA[numCameras + 1];
CAMERA LastCam;

// Decals
int maxDecals = 100;
DECAL muzzleflashdecal;
DECAL impactflashdecal;

// Lighting
LIGHT* light = new LIGHT[numLights + 1];
float lightColor[3] = {1.0, 1.0, 1.0};

// Dialogs
HWND hWndStartDlg;
int DialogInUse = 0;
char szText[BUFFER_SIZE];

// Font
GLFONT myFont;

// Bounding box variables
float Min_X, Min_Y, Min_Z, Max_X, Max_Y, Max_Z;

// Test variables
int numleavesvisible = 0;
int showportals = 0;
int numlistleaves = 0;
int numlistpartitions = 0;
int numcurrentportals = 0;

// Sound
int device;
BOOL lowqual;
int numSamples = 0;
int numChannels = 0;
LinkedList<SOUND_SAMPLE> SampleList;
LinkedList<SOUND_CHANNEL> ChannelList;

// Bullet
int bulletspeed = 0;
int bullettype = 1;
int numBulletsActive = 0;
int numBullets = 50;
float shotlast;
BULLET* bullet = new BULLET[numBullets];

// BSP
int numleaves = 0;
int currentleaf = 0;
int numpartitions = 0;
bool BuiltBSP = false;
BSP_node* root = new BSP_node;

// Brush class
int numBrushes = 3;                  // number of brushes
POLYGON* brushPolygons = new POLYGON[14];  // polygons for brush initialization
BRUSH* BrushSet = NULL;                    // pointer to a list of brushes
BRUSH* Brush = new BRUSH[3];      // original brushes
VERTEX* vertex1 = new VERTEX[8];           // Vertices for the brushes
VERTEX* vertex2 = new VERTEX[8];
VERTEX* vertex3 = new VERTEX[8];

// Portals
int numportals = 0;
LinkedList<PORTAL> PortalList;
LinkedList<PORTAL> BrushPortalList;
PORTAL* portal;

// Static Lights
int numStaticLights = 9;
StaticLight* staticlight = new StaticLight[numStaticLights];

// Lightmap
int numLightmaps = 0;
Lightmap* lightmap = new Lightmap[numWorldPolygons];

// Splines
int visible = 0;                // Toggle flag for viewing the splines
int numSplines = 0;             // Not zero based
int cameraMode = 2;             // Camera mode (0 = Free, 1 = Look, 2 = Follow)
int currentSpline = 0;          // Current spline to work with or use for a path
int lookAtPath = 0;             // Spline to look at when in Follow mode
char SplineFileName[MAX_PATH];  // filename and path to the spline data
SPLINE* spline;                 // Global pointer for working with the splines from list
LinkedList<SPLINE> SplineList;  // Spline list. Note that you can't make this a pointer.

// Linked Lists of leaf, partition and portal nodes
LinkedList<BSP_node> LeafList;
LinkedList<BSP_node> PartitionList;
BSP_node* listnode;

// Particles
ParticleManager PManager; // instance of particle manager

// Rendering state
bool Texturing = 0;
bool CSGflag = 0;
bool CSGcullface = 1;

int InitGL(int Width, int Height)
{
    int ReturnValue;
    // Seed the random number generator
    srand((unsigned)time(NULL));
    // Initialize the sound
    InitializeBASS();
    CreateSounds();
    // Set the application start time
    ApplicationStartTime = (float)GetTickCount();
    // Set the initial camera mode
    cameraMode = 0;
    // Set the OpenGL state
    SetGLProperties();
    SetGLMaterial();
    SetGLLighting(light);
    // Set the camera and splines
    SetGLCamera(camera);
    SetSplines(SplineList);
    // Set the textures
    if (!SetGLTexture(texture))
        ReturnValue = 0;
    else
        ReturnValue = 1;

    // Create the font texture
    glFontCreate(&myFont, "roman.glf", 600);

    // Initialize the world geometry
    SetGLWorld(worldPolygons, texture, vertex);

    // Find the bounding box of the data set
    Min_X = worldPolygons[0].Vertex[0].coords.x;
    Min_Y = worldPolygons[0].Vertex[0].coords.y;
    Min_Z = worldPolygons[0].Vertex[0].coords.z;
    Max_X = worldPolygons[0].Vertex[0].coords.x;
    Max_Y = worldPolygons[0].Vertex[0].coords.y;
    Max_Z = worldPolygons[0].Vertex[0].coords.z;
    for (int loop = 0; loop < numWorldPolygons; loop++)
    {
        for (int i = 0; i < 3; i++)
        {
            if (worldPolygons[loop].Vertex[i].coords.x < Min_X )
                Min_X = worldPolygons[loop].Vertex[i].coords.x;
            if (worldPolygons[loop].Vertex[i].coords.y < Min_Y )
                Min_Y = worldPolygons[loop].Vertex[i].coords.y;
            if (worldPolygons[loop].Vertex[i].coords.z < Min_Z )
                Min_Z = worldPolygons[loop].Vertex[i].coords.z;
            if (worldPolygons[loop].Vertex[i].coords.x > Max_X )
                Max_X = worldPolygons[loop].Vertex[i].coords.x;
            if (worldPolygons[loop].Vertex[i].coords.y > Max_Y )
                Max_Y = worldPolygons[loop].Vertex[i].coords.y;
            if (worldPolygons[loop].Vertex[i].coords.z > Max_Z )
                Max_Z = worldPolygons[loop].Vertex[i].coords.z;
        }
    }

    // BSP initialization
    for (int i = 0; i < numWorldPolygons; i++)
        bspPolygons[i] = worldPolygons[i];

    root->nodeid = 0;
    root->leaf = 0;
    root->visible = 0;
    root->numpolys = numWorldPolygons;
    root->nodepolylist = bspPolygons;
    root->nodelightmaplist = lightmap;
    root->numdecals = 0;
    BuildBSP(root);
    BuiltBSP = true;

    // Portal Creation
    MakeNodeLists(root);
    MakePortalList();
    AddPortalsToLeaves(root);
    FindTruePortals(root);

    // Initialize static lights
    SetStaticLights(staticlight);

    // Create the lightmaps
    CreateBSPLightmaps(root);
    LoadBSPLightmaps(root);

    // Model Initialization
    InitializeBullets();

    // Decal Initialization
    muzzleflashdecal.Size = 3;
    impactflashdecal.Size = 3;

/*
    // Particle Initialization
    VECTOR tempvect(0.0, -10.0, -30);
    CreateRomanCandle(tempvect);
    tempvect.Set(-10.0, -10.0, -30);
    CreateBouncy(tempvect);
    tempvect.Set(10.0, -10.0, -30);
    CreateBouncy(tempvect);
//*/

    // CSG Brush Initialization
    ResetBrushes();

    return ReturnValue;
}

void ReSizeGLScene(int Width, int Height)
{
    SetGLView(Width, Height);
}

void DrawGLScene(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    LastCam = camera[currentCamera];
    camera[currentCamera].Update();

    CollisionPacket cp;
    cp.eRadius.x = 1.0;
    cp.eRadius.y = 1.0;
    cp.eRadius.z = 1.0;
    cp.sourcePoint.x = LastCam.Position.x;
    cp.sourcePoint.y = LastCam.Position.y;
    cp.sourcePoint.z = LastCam.Position.z;
    cp.velocity.x = camera[currentCamera].Position.x - LastCam.Position.x;
    cp.velocity.y = camera[currentCamera].Position.y - LastCam.Position.y;
    cp.velocity.z = camera[currentCamera].Position.z - LastCam.Position.z;

    CheckForCollision(&camera[currentCamera].Position, numWorldPolygons, worldPolygons, &cp);

    camera[currentCamera].Apply();

    for(int loop = 0; loop <= numLights; loop++)
        light[loop].Apply();

    currentleaf = FindCurrentLeaf(camera[currentCamera].Position, root);

    ExtractFrustum();
    CalculatePVS(currentleaf);
    numleavesvisible = CountVisibleLeaves();

    UpdateListener();
    UpdateChannel(channel_gunshot, &camera[currentCamera].Position, NULL, NULL);

    UpdateBullets();

    PManager.Update();

    DrawWorld(root);

    DrawDecals();

    // Draw CSG Brushes
    // The function can be found at the end of general.cpp
    DrawBrushes();

    DrawSplines(SplineList);           // Required to calculate the camera path

    DrawMyText();              // Draw text last
}

LRESULT CALLBACK WndProc(    HWND    hWnd,
                            UINT    message,
                            WPARAM    wParam,
                            LPARAM    lParam)
{
    GLuint    PixelFormat;
    int generic_format;
    int generic_accelerated;
    PIXELFORMATDESCRIPTOR pfd_new;
  
    static PIXELFORMATDESCRIPTOR pfd=
    {
           sizeof(PIXELFORMATDESCRIPTOR),
        1,
        PFD_DRAW_TO_WINDOW |
        PFD_SUPPORT_OPENGL |
        PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA,
        16,
        0, 0, 0, 0, 0, 0,
        0,
        0,
        0,
        0, 0, 0, 0,
        16,
        0,
        0,
        PFD_MAIN_PLANE,
        0,
        0, 0, 0
    };


    switch (message)
    {
        case WM_CREATE:
            hDC = GetDC(hWnd);

            PixelFormat = ChoosePixelFormat(hDC, &pfd);

             if (!PixelFormat)
            {
                   MessageBox(NULL,"Can't find a suitable PixelFormat.","Error",MB_OK|MB_ICONERROR);
                PostQuitMessage(0);
                break;
            }

            if(!SetPixelFormat(hDC,PixelFormat,&pfd))
            {
                MessageBox(NULL,"Can't set the PixelFormat.","Error",MB_OK|MB_ICONERROR);
                PostQuitMessage(0);
                break;
            }

            hRC = wglCreateContext(hDC);
            if(!hRC)
            {
                MessageBox(NULL,"Can't create a GL Rendering Context.","Error",MB_OK|MB_ICONERROR);
                PostQuitMessage(0);
                break;
            }

            if(!wglMakeCurrent(hDC, hRC))
            {
                MessageBox(NULL,"Can't activate the GL Rendering Context.","Error",MB_OK|MB_ICONERROR);
                PostQuitMessage(0);
                break;
            }
            break;

        case WM_SYSCOMMAND:
        {
            switch (wParam)
            {
                case SC_SCREENSAVE:
                case SC_MONITORPOWER:
                    return 0;
            }
            break;
        }

        if (DialogInUse)
        {
            case WM_PAINT:
                BeginPaint(hWnd,&ps);
                DrawGLScene();
                glFlush();
                SwapBuffers(hDC);
                EndPaint(hWnd,&ps);
            break;
        }

        case WM_CLOSE:
            delete[] lightmap;
            delete[] staticlight;
            delete[] texture;
            delete[] vertex;
            delete[] camera;
            delete[] light;
            delete[] worldPolygons;

            for (int i = numSplines - 1; i >= 0; i--)
                DeleteSpline(i, SplineList);

// TODO: Fix This
//            if (BuiltBSP)
//                DeleteBSP(root);
//            delete root;

            for (int i = numportals; i > 0; i--)
            {
                portal = PortalList.Get(i);
                delete[] portal->Vertex;
            }
           
            glFontDestroy(&myFont);
            
            BASS_Stop();             // Stop all samples playing
            BASS_Free();             // Free the BASS resources

            DeleteBullets();
            
            CLog::addLine("--- End of Program ---");
            Console.Close();

            ChangeDisplaySettings(NULL, 0);
            wglMakeCurrent(hDC,NULL);
            wglDeleteContext(hRC);
            ReleaseDC(hWnd,hDC);
            DestroyWindow(hWnd);
        break;

        case WM_DESTROY:
            PostQuitMessage(0);
        break;

        case WM_KEYDOWN:
            key[wParam] = TRUE;
        break;

        case WM_KEYUP:
            key[wParam] = FALSE;
        break;

        case WM_SIZE:
            SetCursorPos((int)(screen.right * 0.5), (int)(screen.bottom * 0.5));
            ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));
        break;

        case WM_MOUSEMOVE:
            if (cameraMode == 0)
            {
                camera[currentCamera].Delta_x = float(HIWORD(lParam) - screen.bottom * 0.5) * 10;
                camera[currentCamera].Delta_y = float(LOWORD(lParam) - screen.right * 0.5) * 10;
            }
        break;

        case WM_LBUTTONDOWN:
            camera[currentCamera].Delta_z = -120.0;
        break;

        case WM_RBUTTONDOWN:
            camera[currentCamera].Delta_z = 120.0;
        break;

        case WM_LBUTTONUP:
            if (wParam != MK_RBUTTON)
                camera[currentCamera].Delta_z = 0.0;
        break;

        case WM_RBUTTONUP:
            if (wParam != MK_LBUTTON)
                camera[currentCamera].Delta_z = 0.0;
        break;

        default:
            return (DefWindowProc(hWnd, message, wParam, lParam));
    }
    return (0);
}

int WINAPI WinMain( HINSTANCE    hInstance,
                    HINSTANCE,
                    LPSTR,
                    int)
{
    MSG        msg;
    g_hInst = hInstance;
    GetWindowRect(GetDesktopWindow(), &screen);

    WNDCLASSEX    wc;
    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style            = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS | CS_SAVEBITS;
    wc.lpfnWndProc        = (WNDPROC) WndProc;
    wc.cbClsExtra        = 0;
    wc.cbWndExtra        = 0;
    wc.hInstance        = hInstance;
    wc.hIcon            = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MYICON));
    wc.hIconSm          = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MYICON));
    wc.hCursor            = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground    = NULL;
    wc.lpszMenuName        = NULL;
    wc.lpszClassName    = "OpenGL WinClass";


    if(!RegisterClassEx(&wc))
    {
        MessageBox(NULL,"Failed To Register The Window Class.","Error",MB_OK|MB_ICONERROR);
        return FALSE;
    }

    hWnd = CreateWindowEx(
    WS_EX_LEFT,
    "OpenGL WinClass",
    "OpenGL & Win32 Tutorial No.20",
    WS_MAXIMIZE |
    WS_CLIPCHILDREN |
    WS_CLIPSIBLINGS |
    WS_POPUPWINDOW |
    WS_VISIBLE,
    0, 0,
    screen.right, screen.bottom,
    NULL,
    NULL,
    hInstance,
    NULL);

    if(!hWnd)
    {
        MessageBox(NULL,"Window Creation Error.","Error",MB_OK|MB_ICONERROR);
        return FALSE;
    }

    DEVMODE dmScreenSettings;
    memset(&dmScreenSettings, 0, sizeof(DEVMODE));
    dmScreenSettings.dmSize        = sizeof(DEVMODE);
    dmScreenSettings.dmPelsWidth    = screen.right;
    dmScreenSettings.dmPelsHeight    = screen.bottom;
    dmScreenSettings.dmFields    = DM_PELSWIDTH | DM_PELSHEIGHT;
    ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN);

    ShowWindow(hWnd, SW_HIDE);

    GetCurrentDirectory(MAX_PATH, AppDirectory);

    DialogBox(hInstance, "DEVICE", hWnd, (DLGPROC)devicedialogproc);

	Console.Open();

    if(DialogBox(hInstance, "STARTDLG", hWnd, (DLGPROC)StartProc) == IDOK)
    {
        if (!InitGL(screen.right, screen.bottom))
        {
            SendMessage(hWnd, WM_CLOSE, 0, 0);
        }
        else
        {
            ShowWindow(hWnd, SW_SHOW);
            UpdateWindow(hWnd);
            SetFocus(hWnd);
            wglMakeCurrent(hDC,hRC);
            SetCursorPos((int)(screen.right * 0.5), (int)(screen.bottom * 0.5));
            ShowCursor(0);
        }
    }
    else
    {
        delete[] brushPolygons;
        delete[] Brush;
        delete[] vertex1;
        delete[] vertex2;
        delete[] vertex3;
        delete[] lightmap;
        delete[] staticlight;
        delete[] texture;
        delete[] vertex;
        delete[] camera;
        delete[] light;
        delete[] worldPolygons;
        delete[] bspPolygons;
        delete root;
        delete[] bullet;
        PostQuitMessage(0);
    }

    while (1)
    {
        while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (GetMessage(&msg, NULL, 0, 0))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
            else
            {
                return TRUE;
            }
        }

        if(!DialogInUse)
        {
            SetCursorPos((int)(screen.right * 0.5), (int)(screen.bottom * 0.5));
            DrawGLScene();
            glFlush();
            SwapBuffers(hDC);
        }

        multiplier = GetTimePassed(lasttime, average, lastmultiplier);
        camera[currentCamera].Multiplier = multiplier;
        light[currentLight].Multiplier = multiplier;
        for (int loop = 0; loop < numBullets; loop++)
            bullet[loop].Multiplier = multiplier;

        if (key['A'] && released_key['A'] == 0 && CSGflag == 1)
        {
            CSGAddition();
            released_key['A'] = 1;
        }

        if (!key['A'])
            released_key['A'] = 0;

        if (key['T'] && released_key['T'] == 0)
        {
            Texturing = !Texturing;
            released_key['T'] = 1;
        }

        if (!key['T'])
            released_key['T'] = 0;

        if (key['C'] && released_key['C'] == 0)
        {
            CSGcullface = !CSGcullface;
            released_key['C'] = 1;
        }

        if (!key['C'])
            released_key['C'] = 0;

        if (key['R'] && released_key['R'] == 0)
        {
            ResetBrushes();
            released_key['R'] = 1;
        }

        if (!key['R'])
            released_key['R'] = 0;

        if (key['M'] && released_key['M'] == 0)
        {
            if (cameraMode == 2)
                cameraMode = 0;
            else
                cameraMode = 2;
            released_key['M'] = 1;
        }

        if (!key['M'])
            released_key['M'] = 0;

        if (key['F'])
        {
            // Loop through all bullets
            for (int loop = 0; loop < numBullets; loop++)
            {
                // Get current time
                float shotthis = (float)GetTickCount();
                // If last time in NULL then this is the first time fired
                if (shotlast == 0)
                    shotlast = ApplicationStartTime;
                // Get the interval between shots
                float shotoffset = shotthis - shotlast;
                // If minimum interval has passed and bullet is inactive
                if (shotoffset > 100 && bullet[loop].active != true)
                {
                    // This time becomes last time fired
                    shotlast = shotthis;
                    // Set a flag to indicate that this is the first time fired
                    bullet[loop].flag = true;
                    // Set the bullet to active
                    bullet[loop].active = true;
                    // Set the bullets direction vector
                    bullet[loop].VelocityVector = camera[currentCamera].GetZUnit();
                    // Set the bullets orientation
                    bullet[loop].Orientation = camera[currentCamera].Orientation;
                    // Set the bullets position
                    bullet[loop].OriginalPosition = camera[currentCamera].Position;
                    bullet[loop].PreviousPosition = camera[currentCamera].Position;
                    bullet[loop].Position = camera[currentCamera].Position;
                    // Increase the number of bullets active counter
                    numBulletsActive++;
                    // Make the muzzle-flash decal active
                    muzzleflashdecal.active = true;
                    // Play the gunshot sound
                    PlayChannel(channel_gunshot);

                    break;
                }
            }
        }

        if (key['1'])
        {
            step = 10.0;
        }

        if (key['2'])
        {
            step = 20.0;
        }

        if (key['3'])
        {
            step = 30.0;
        }

        if (key['4'])
        {
            step = 40.0;
        }

        if (key['5'])
        {
            step = 50.0;
        }

        if (key['6'])
        {
            step = 60.0;
        }

        if (key['7'])
        {
            step = 70.0;
        }

        if (key['8'])
        {
            step = 80.0;
        }

        if (key['9'])
        {
            step = 90.0;
        }

        if (key['0'])
        {
            step = 100.0;
        }

        if (key[VK_NUMPAD6])
        {
            light[currentLight].Movement_x += step;
        }

        if (key[VK_NUMPAD4])
        {
            light[currentLight].Movement_x -= step;
        }

        if (key[VK_NUMPAD2])
        {
            light[currentLight].Movement_z += step;
        }

        if (key[VK_NUMPAD8])
        {
            light[currentLight].Movement_z -= step;
        }

        if (key[VK_NUMPAD7])
        {
            light[currentLight].Movement_y += step;
        }

        if (key[VK_NUMPAD9])
        {
            light[currentLight].Movement_y -= step;
        }

        if (key[VK_NUMPAD5] && released_key[VK_NUMPAD5] == 0)
        {
            if (light[currentLight].Positional == FALSE)
                light[currentLight].Positional = TRUE;
            else
                light[currentLight].Positional = FALSE;
            released_key[VK_NUMPAD5] = 1;
        }

        if (!key[VK_NUMPAD5])
            released_key[VK_NUMPAD5] = 0;

        if (key[VK_RIGHT])
        {
            camera[currentCamera].Movement_x += step;
        }

        if (key[VK_LEFT])
        {
            camera[currentCamera].Movement_x -= step;
        }

        if (key[VK_DOWN])
        {
            camera[currentCamera].Movement_z += step;
        }

        if (key[VK_UP])
        {
            camera[currentCamera].Movement_z -= step;
        }

        if (key[VK_PRIOR])
        {
            camera[currentCamera].Movement_y += step;
        }

        if (key[VK_NEXT])
        {
            camera[currentCamera].Movement_y -= step;
        }

        if (key[VK_SPACE])
        {
            camera[currentCamera].Reset();
        }

        if (key[VK_ESCAPE] || key['Q'])
            SendMessage(hWnd,WM_CLOSE,0,0);
    }
}



                                      



